"""
Creates and manages isolated build environments.
"""
import abc
import os
import platform
import shutil
import subprocess
import sys
import sysconfig
import tempfile
from types import TracebackType
from typing import Iterable, Optional, Tuple, Type

from ._compat import abstractproperty, add_metaclass

try:
    import pip
except ImportError:  # pragma: no cover
    pip = None  # pragma: no cover


@add_metaclass(abc.ABCMeta)
class IsolatedEnv(object):
    """Abstract base of isolated build environments, as required by the build project."""

    @abstractproperty
    def executable(self):  # type: () -> str
        """Return the executable of the isolated build environment."""
        raise NotImplementedError

    @abc.abstractmethod
    def install(self, requirements):  # type: (Iterable[str]) -> None
        """
        Install PEP-508 requirements into the isolated build environment.

        :param requirements: PEP-508 requirements
        """
        raise NotImplementedError


class IsolatedEnvBuilder(object):
    def __init__(self):  # type: () -> None
        """Builder object for isolated environment."""
        self._path = None  # type: Optional[str]

    def __enter__(self):  # type: () -> IsolatedEnv
        """
        Creates an isolated build environment.

        :return: the isolated build environment
        """
        self._path = tempfile.mkdtemp(prefix='build-env-')
        try:
            executable, pip_executable = _create_isolated_env(self._path)
            return _IsolatedEnvVenvPip(path=self._path, python_executable=executable, pip_executable=pip_executable)
        except Exception:  # cleanup folder if creation fails
            self.__exit__(*sys.exc_info())
            raise

    def __exit__(self, exc_type, exc_val, exc_tb):
        # type: (Optional[Type[BaseException]], Optional[BaseException], Optional[TracebackType]) -> None
        """
        Delete the created isolated build environment.

        :param exc_type: the type of exception raised (if any)
        :param exc_val: the value of exception raised (if any)
        :param exc_tb: the traceback of exception raised (if any)
        """
        if self._path is not None and os.path.exists(self._path):  # in case the user already deleted skip remove
            shutil.rmtree(self._path)


class _IsolatedEnvVenvPip(IsolatedEnv):
    """
    Isolated build environment context manager

    Non-standard paths injected directly to sys.path still be passed to the environment.
    """

    def __init__(self, path, python_executable, pip_executable):
        # type: (str, str, str) -> None
        """
        Define an isolated build environment.

        :param path: the path where the environment exists
        :param python_executable: the python executable within the environment
        :param pip_executable: an executable that allows installing packages within the environment
        """
        self._path = path
        self._pip_executable = pip_executable
        self._python_executable = python_executable

    @property
    def path(self):  # type: () -> str
        """:return: the location of the isolated build environment"""
        return self._path

    @property
    def executable(self):  # type: () -> str
        """:return: the python executable of the isolated build environment"""
        return self._python_executable

    def install(self, requirements):  # type: (Iterable[str]) -> None
        """
        Installs the specified PEP 508 requirements on the environment

        :param requirements: PEP-508 requirement specification to install

        :note: Passing non PEP 508 strings will result in undefined behavior, you *should not* rely on it. It is \
               merely an implementation detail, it may change any time without warning.
        """
        if not requirements:
            return

        with tempfile.NamedTemporaryFile('w+', prefix='build-reqs-', suffix='.txt', delete=False) as req_file:
            req_file.write(os.linesep.join(requirements))
        try:
            cmd = [
                self._pip_executable,
                # on python2 if isolation is achieved via environment variables, we need to ignore those while calling
                # host python (otherwise pip would not be available within it)
                '-{}m'.format('E' if self._pip_executable == self.executable and sys.version_info[0] == 2 else ''),
                'pip',
                'install',
                '--prefix',
                self.path,
                '--ignore-installed',
                '--no-warn-script-location',
                '-r',
                os.path.abspath(req_file.name),
            ]
            subprocess.check_call(cmd)
        finally:
            os.unlink(req_file.name)


if sys.version_info[0] == 2:  # noqa: C901 # disable if too complex

    def _create_isolated_env(path):  # type: (str) -> Tuple[str, str]
        """
        On Python 2 we use the virtualenv package to provision a virtual environment.

        :param path: the folder where to create the isolated build environment
        :return: the isolated build environment executable, and the pip to use to install packages into it
        """
        from virtualenv import cli_run

        cmd = [str(path), '--no-setuptools', '--no-wheel', '--activators', '']
        if pip is not None:
            cmd.append('--no-pip')
        result = cli_run(cmd, setup_logging=False)
        executable = str(result.creator.exe)
        pip_executable = executable if pip is None else sys.executable
        return executable, pip_executable


else:

    def _create_isolated_env(path):  # type: (str) -> Tuple[str, str]
        """
        On Python 3 we use the venv package from the standard library, and if host python has no pip the ensurepip
        package to provision one into the created virtual environment.

        :param path: the folder where to create the isolated build environment
        :return: the isolated build environment executable, and the pip to use to install packages into it
        """
        import venv

        venv.EnvBuilder(with_pip=False).create(path)
        executable = _find_executable(path)

        # Scenario 1: pip is available (either installed or via pth file) within the python executable alongside
        # this projects environment: in this case we should be able to import it
        if pip is not None:
            pip_executable = sys.executable
        else:
            # Scenario 2: this project is installed into a virtual environment that has no pip, but the system has
            # Scenario 3: there's a pip executable on PATH
            # Scenario 4: no pip can be found, we might be able to provision one into the build env via ensurepip
            cmd = [executable, '-Im', 'ensurepip', '--upgrade', '--default-pip']
            try:
                subprocess.check_call(cmd, cwd=path)
            except subprocess.CalledProcessError:  # pragma: no cover
                pass  # pragma: no cover
            # avoid the setuptools from ensurepip to break the isolation
            subprocess.check_call([executable, '-Im', 'pip', 'uninstall', 'setuptools', '-y'])
            pip_executable = executable
        return executable, pip_executable

    def _find_executable(path):  # type: (str) -> str
        """
        Detect the executable within a virtual environment.

        :param path: the location of the virtual environment
        :return: the python executable
        """
        config_vars = sysconfig.get_config_vars().copy()  # globally cached, copy before altering it
        config_vars['base'] = path
        env_scripts = sysconfig.get_path('scripts', vars=config_vars)
        if not env_scripts:
            raise RuntimeError("Couldn't get environment scripts path")
        exe = 'pypy3' if platform.python_implementation() == 'PyPy' else 'python'
        if os.name == 'nt':
            exe = '{}.exe'.format(exe)
        executable = os.path.join(path, env_scripts, exe)
        if not os.path.exists(executable):
            raise RuntimeError('Virtual environment creation failed, executable {} missing'.format(executable))
        return executable


__all__ = (
    'IsolatedEnvBuilder',
    'IsolatedEnv',
)



# SPDX-License-Identifier: MIT

from __future__ import unicode_literals

import copy
import os
import sys
import typing

import pep517.wrappers
import pytest

import build

if sys.version_info >= (3, 8):  # pragma: no cover
    import email
    from importlib import metadata as importlib_metadata

    email_message_from_string = email.message_from_string
else:  # pragma: no cover
    import importlib_metadata

    email_message_from_string = importlib_metadata._compat.email_message_from_string

if sys.version_info >= (3,):  # pragma: no cover
    build_open_owner = 'builtins'
else:  # pragma: no cover
    build_open_owner = 'build'
    FileNotFoundError = IOError
    PermissionError = OSError


DUMMY_METADATA = """
Version: 1.0.0
Provides-Extra: some_extra
""".strip()


DEFAULT_BACKEND = {
    'build-backend': 'setuptools.build_meta:__legacy__',
    'requires': ['setuptools >= 40.8.0', 'wheel'],
}


class MockDistribution(importlib_metadata.Distribution):
    def read_text(self, filename):
        return DUMMY_METADATA

    def locate_file(self, path):  # pragma: no cover
        return ''

    @classmethod
    def from_name(cls, name):
        if name == 'something':
            return cls()
        raise importlib_metadata.PackageNotFoundError


importlib_metadata.Distribution = MockDistribution


@pytest.mark.parametrize(
    ('requirement_string', 'expected', 'extra_warning'),
    [
        ('something', True, False),
        ('something_else', False, False),
        ('something[extra]', False, False),
        ('something[some_extra]', True, True),
        ('something_else; python_version>"10"', True, False),
        ('something_else; python_version<="1"', True, False),
        ('something_else; python_version>="1"', False, False),
        ('something == 1.0.0', True, False),
        ('something == 2.0.0', False, False),
        ('something[some_extra] == 1.0.0', True, True),
        ('something[some_extra] == 2.0.0', False, True),
    ],
)
def test_check_version(requirement_string, expected, extra_warning):
    if extra_warning:
        with pytest.warns(build.IncompleteCheckWarning):
            assert build.check_version(requirement_string) == expected
    else:
        assert build.check_version(requirement_string) == expected


def test_init(mocker, test_flit_path, legacy_path, test_no_permission, test_bad_syntax_path):
    modules = {
        'flit_core.buildapi': None,
        'setuptools.build_meta:__legacy__': None,
    }
    mocker.patch('importlib.import_module', modules.get)
    mocker.patch('pep517.wrappers.Pep517HookCaller')

    # correct flit pyproject.toml
    build.ProjectBuilder(test_flit_path)
    pep517.wrappers.Pep517HookCaller.assert_called_with(
        test_flit_path,
        'flit_core.buildapi',
        backend_path=None,
        python_executable=sys.executable,
    )
    pep517.wrappers.Pep517HookCaller.reset_mock()

    # custom python
    build.ProjectBuilder(test_flit_path, python_executable='some-python')
    pep517.wrappers.Pep517HookCaller.assert_called_with(
        test_flit_path,
        'flit_core.buildapi',
        backend_path=None,
        python_executable='some-python',
    )
    pep517.wrappers.Pep517HookCaller.reset_mock()

    # FileNotFoundError
    build.ProjectBuilder(legacy_path)
    pep517.wrappers.Pep517HookCaller.assert_called_with(
        legacy_path,
        'setuptools.build_meta:__legacy__',
        backend_path=None,
        python_executable=sys.executable,
    )

    # PermissionError
    if sys.version_info[0] != 2 and os.name != 'nt':  # can't correctly set the permissions required for this
        with pytest.raises(build.BuildException):
            build.ProjectBuilder(test_no_permission)

    # TomlDecodeError
    with pytest.raises(build.BuildException):
        build.ProjectBuilder(test_bad_syntax_path)


@pytest.mark.parametrize('value', [b'something', typing.Text('something_else')])
def test_python_executable(test_flit_path, value):
    builder = build.ProjectBuilder(test_flit_path)

    builder.python_executable = value
    assert builder.python_executable == value
    assert builder._hook.python_executable == value


@pytest.mark.parametrize('distribution', ['wheel', 'sdist'])
def test_get_dependencies_missing_backend(packages_path, distribution):
    bad_backend_path = os.path.join(packages_path, 'test-bad-backend')
    builder = build.ProjectBuilder(bad_backend_path)

    with pytest.raises(build.BuildException):
        builder.get_dependencies(distribution)


@pytest.mark.parametrize('distribution', ['wheel', 'sdist'])
def test_build_missing_backend(packages_path, distribution, tmpdir):
    bad_backend_path = os.path.join(packages_path, 'test-bad-backend')
    builder = build.ProjectBuilder(bad_backend_path)

    with pytest.raises(build.BuildException):
        builder.build(distribution, str(tmpdir))


def test_check_dependencies(mocker, test_flit_path):
    mocker.patch('importlib.import_module', autospec=True)
    mocker.patch('pep517.wrappers.Pep517HookCaller.get_requires_for_build_sdist')
    mocker.patch('pep517.wrappers.Pep517HookCaller.get_requires_for_build_wheel')
    mocker.patch('build.check_version', autospec=True)

    builder = build.ProjectBuilder(test_flit_path)

    side_effects = [
        [],
        ['something'],
        pep517.wrappers.BackendUnavailable,
    ]

    build.check_version.return_value = False
    builder._hook.get_requires_for_build_sdist.side_effect = copy.copy(side_effects)
    builder._hook.get_requires_for_build_wheel.side_effect = copy.copy(side_effects)

    # requires = []
    assert builder.check_dependencies('sdist') == {'flit_core >=2,<3'}
    assert builder.check_dependencies('wheel') == {'flit_core >=2,<3'}

    # requires = ['something']
    assert builder.check_dependencies('sdist') == {'flit_core >=2,<3', 'something'}
    assert builder.check_dependencies('wheel') == {'flit_core >=2,<3', 'something'}

    # BackendUnavailable
    with pytest.raises(build.BuildBackendException):
        builder.check_dependencies('sdist')
    with pytest.raises(build.BuildBackendException):
        not builder.check_dependencies('wheel')


def test_working_directory(tmp_dir):
    assert os.path.realpath(os.curdir) != os.path.realpath(tmp_dir)
    with build._working_directory(tmp_dir):
        assert os.path.realpath(os.curdir) == os.path.realpath(tmp_dir)


def test_build(mocker, test_flit_path, tmp_dir):
    mocker.patch('importlib.import_module', autospec=True)
    mocker.patch('pep517.wrappers.Pep517HookCaller', autospec=True)
    mocker.patch('build._working_directory', autospec=True)

    builder = build.ProjectBuilder(test_flit_path)

    builder._hook.build_sdist.side_effect = [None, Exception]
    builder._hook.build_wheel.side_effect = [None, Exception]

    builder.build('sdist', tmp_dir)
    builder._hook.build_sdist.assert_called_with(tmp_dir, {})
    build._working_directory.assert_called_with(test_flit_path)

    builder.build('wheel', tmp_dir)
    builder._hook.build_wheel.assert_called_with(tmp_dir, {})
    build._working_directory.assert_called_with(test_flit_path)

    with pytest.raises(build.BuildBackendException):
        build._working_directory.assert_called_with(test_flit_path)
        builder.build('sdist', tmp_dir)

    with pytest.raises(build.BuildBackendException):
        build._working_directory.assert_called_with(test_flit_path)
        builder.build('wheel', tmp_dir)


def test_default_backend(mocker, legacy_path):
    mocker.patch('importlib.import_module', autospec=True)
    mocker.patch('pep517.wrappers.Pep517HookCaller', autospec=True)

    builder = build.ProjectBuilder(legacy_path)

    assert builder._build_system == DEFAULT_BACKEND


def test_missing_backend(mocker, test_no_backend_path):
    mocker.patch('importlib.import_module', autospec=True)
    mocker.patch('pep517.wrappers.Pep517HookCaller', autospec=True)

    builder = build.ProjectBuilder(test_no_backend_path)

    assert builder._build_system == DEFAULT_BACKEND


def test_missing_requires(mocker, test_no_requires_path):
    mocker.patch('importlib.import_module', autospec=True)
    mocker.patch('pep517.wrappers.Pep517HookCaller', autospec=True)

    with pytest.raises(build.BuildException):
        build.ProjectBuilder(test_no_requires_path)


def test_build_system_typo(mocker, test_typo):
    mocker.patch('importlib.import_module', autospec=True)
    mocker.patch('pep517.wrappers.Pep517HookCaller', autospec=True)

    with pytest.warns(build.TypoWarning):
        build.ProjectBuilder(test_typo)


def test_missing_outdir(mocker, tmp_dir, test_flit_path):
    mocker.patch('importlib.import_module', autospec=True)
    mocker.patch('pep517.wrappers.Pep517HookCaller', autospec=True)

    builder = build.ProjectBuilder(test_flit_path)
    out = os.path.join(tmp_dir, 'out')

    builder.build('sdist', out)

    assert os.path.isdir(out)


def test_relative_outdir(mocker, tmp_dir, test_flit_path):
    mocker.patch('importlib.import_module', autospec=True)
    mocker.patch('pep517.wrappers.Pep517HookCaller', autospec=True)

    builder = build.ProjectBuilder(test_flit_path)

    builder.build('sdist', '.')

    builder._hook.build_sdist.assert_called_with(os.path.abspath('.'), {})


def test_not_dir_outdir(mocker, tmp_dir, test_flit_path):
    mocker.patch('importlib.import_module', autospec=True)
    mocker.patch('pep517.wrappers.Pep517HookCaller', autospec=True)

    builder = build.ProjectBuilder(test_flit_path)
    out = os.path.join(tmp_dir, 'out')

    open(out, 'a').close()  # create empty file

    with pytest.raises(build.BuildException):
        builder.build('sdist', out)



# SPDX-License-Identifier: MIT

import argparse
import os
import sys
import traceback
import warnings

from typing import List, Optional, TextIO, Type, Union, Iterable

import build
from build import BuildBackendException, BuildException, ConfigSettings, ProjectBuilder
from build.env import IsolatedEnvBuilder

__all__ = ['build', 'main', 'main_parser']


def _showwarning(message, category, filename, lineno, file=None, line=None):  # pragma: no cover
    # type: (Union[Warning, str], Type[Warning], str, int, Optional[TextIO], Optional[str]) -> None
    prefix = 'WARNING'
    if sys.stdout.isatty():
        prefix = '\33[93m' + prefix + '\33[0m'
    print('{} {}'.format(prefix, str(message)))


warnings.showwarning = _showwarning


def _error(msg, code=1):  # type: (str, int) -> None  # pragma: no cover
    """
    Prints an error message and exits. Will color the output when writting to a TTY.

    :param msg: Error message
    :param code: Error code
    """
    prefix = 'ERROR'
    if sys.stdout.isatty():
        prefix = '\33[91m' + prefix + '\33[0m'
    print('{} {}'.format(prefix, msg))
    exit(code)


def _build_in_isolated_env(builder, outdir, distributions):
    # type: (ProjectBuilder, str, List[str]) -> None
    with IsolatedEnvBuilder() as env:
        builder.python_executable = env.executable
        env.install(builder.build_dependencies)
        for distribution in distributions:
            builder.build(distribution, outdir)


def _build_in_current_env(builder, outdir, distributions, skip_dependencies=False):
    # type: (ProjectBuilder, str, List[str], bool) -> None
    for dist in distributions:
        if not skip_dependencies:
            missing = builder.check_dependencies(dist)
            if missing:
                _error('Missing dependencies:' + ''.join(['\n\t' + dep for dep in missing]))

        builder.build(dist, outdir)


def build_package(srcdir, outdir, distributions, config_settings=None, isolation=True, skip_dependencies=False):
    # type: (str, str, List[str], Optional[ConfigSettings], bool, bool) -> None
    """
    Runs the build process

    :param srcdir: Source directory
    :param outdir: Output directory
    :param distributions: Distributions to build (sdist and/or wheel)
    :param config_settings: Configuration settings to be passed to the backend
    :param isolation: Isolate the build in a separate environment
    :param skip_dependencies: Do not perform the dependency check
    """
    if not config_settings:
        config_settings = {}

    try:
        builder = ProjectBuilder(srcdir, config_settings)
        if isolation:
            _build_in_isolated_env(builder, outdir, distributions)
        else:
            _build_in_current_env(builder, outdir, distributions, skip_dependencies)
    except BuildException as e:
        _error(str(e))
    except BuildBackendException as e:
        if sys.version_info >= (3, 5):
            print(traceback.format_exc(-1))
        else:
            print(traceback.format_exc())
        _error(str(e))


def main_parser():  # type: () -> argparse.ArgumentParser
    """
    Constructs the main parser
    """
    # mypy does not recognize module.__path__
    # https://github.com/python/mypy/issues/1422
    paths = build.__path__  # type: Iterable[Optional[str]]  # type: ignore
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'srcdir',
        type=str,
        nargs='?',
        default=os.getcwd(),
        help='source directory (defaults to current directory)',
    )
    parser.add_argument(
        '--version',
        '-V',
        action='version',
        version='build {} ({})'.format(build.__version__, ', '.join(path for path in paths if path)),
    )
    parser.add_argument(
        '--sdist',
        '-s',
        action='store_true',
        help='build a source package',
    )
    parser.add_argument(
        '--wheel',
        '-w',
        action='store_true',
        help='build a wheel',
    )
    parser.add_argument(
        '--outdir',
        '-o',
        metavar='dir',
        type=str,
        help='output directory (defaults to {{srcdir}}{sep}dist)'.format(sep=os.sep),
    )
    parser.add_argument(
        '--skip-dependencies',
        '-x',
        action='store_true',
        help='does not check for the dependencies',
    )
    parser.add_argument(
        '--no-isolation',
        '-n',
        action='store_true',
        help='do not isolate the build in a virtual environment',
    )
    parser.add_argument(
        '--config-setting',
        '-C',
        action='append',
        help='pass option to the backend',
    )
    return parser


def main(cli_args, prog=None):  # type: (List[str], Optional[str]) -> None
    """
    Parses the CLI arguments and invokes the build process.

    :param cli_args: CLI arguments
    """
    parser = main_parser()
    if prog:
        parser.prog = prog
    args = parser.parse_args(cli_args)

    distributions = []
    config_settings = {}

    if args.config_setting:
        for arg in args.config_setting:
            if sys.version_info >= (3,):
                data = arg.split('=', maxsplit=1)
            else:
                split_data = arg.split('=')
                data = [
                    split_data[0],
                    '='.join(split_data[1:]),
                ]
            setting = data[0]
            value = data[1] if len(data) >= 2 else ''

            if setting not in config_settings:
                config_settings[setting] = value
            else:
                if not isinstance(config_settings[setting], list):
                    config_settings[setting] = [config_settings[setting]]

                config_settings[setting].append(value)

    if args.sdist:
        distributions.append('sdist')
    if args.wheel:
        distributions.append('wheel')

    # default targets
    if not distributions:
        distributions = ['sdist', 'wheel']

    # outdir is relative to srcdir only if omitted.
    outdir = os.path.join(args.srcdir, 'dist') if args.outdir is None else args.outdir

    build_package(args.srcdir, outdir, distributions, config_settings, not args.no_isolation, args.skip_dependencies)


def entrypoint():  # type: () -> None
    main(sys.argv[1:])


if __name__ == '__main__':  # pragma: no cover
    main(sys.argv[1:], 'python -m build')



# SPDX-License-Identifier: MIT

"""
build - A simple, correct PEP517 package builder
"""
__version__ = '0.1.0'

import contextlib
import difflib
import os
import sys
import warnings

from typing import Iterator, Mapping, Optional, Sequence, Set, Text, Union

import pep517.wrappers
import toml
import toml.decoder

if sys.version_info < (3,):
    FileNotFoundError = IOError
    PermissionError = OSError


ConfigSettings = Mapping[str, Union[str, Sequence[str]]]


_DEFAULT_BACKEND = {
    'build-backend': 'setuptools.build_meta:__legacy__',
    'requires': ['setuptools >= 40.8.0', 'wheel'],
}


class BuildException(Exception):
    """
    Exception raised by ProjectBuilder
    """


class BuildBackendException(Exception):
    """
    Exception raised when the backend fails
    """


class TypoWarning(Warning):
    """
    Warning raised when a potential typo is found
    """


class IncompleteCheckWarning(Warning):
    """
    Warning raised when we have an incomplete check
    """


def check_version(requirement_string, extra=''):  # type: (str, str) -> bool
    """
    :param requirement_string: Requirement string
    :param extra: Extra (eg. test in myproject[test])
    """
    import packaging.requirements

    if sys.version_info >= (3, 8):
        from importlib import metadata as importlib_metadata
    else:
        import importlib_metadata

    req = packaging.requirements.Requirement(requirement_string)
    env = {'extra': extra}

    if req.marker and not req.marker.evaluate(env):
        return True

    try:
        version = importlib_metadata.version(req.name)
        metadata = importlib_metadata.metadata(req.name)
    except importlib_metadata.PackageNotFoundError:
        return False

    for extra in req.extras:
        if extra not in (metadata.get_all('Provides-Extra') or []):
            return False
        warnings.warn(
            "Verified that the '{}[{}]' extra is present but did not verify that it is active "
            '(its dependencies are met)'.format(req.name, extra),
            IncompleteCheckWarning,
        )

    if req.specifier:
        return req.specifier.contains(version)

    return True


def _find_typo(dictionary, expected):  # type: (Mapping[str, str], str) -> None
    if expected not in dictionary:
        for obj in dictionary:
            if difflib.SequenceMatcher(None, expected, obj).ratio() >= 0.8:
                warnings.warn(
                    "Found '{}' in pyproject.toml, did you mean '{}'?".format(obj, expected),
                    TypoWarning,
                )


@contextlib.contextmanager
def _working_directory(path):  # type: (str) -> Iterator[None]
    current = os.getcwd()

    os.chdir(path)

    try:
        yield
    finally:
        os.chdir(current)


class ProjectBuilder(object):
    def __init__(self, srcdir, config_settings=None, python_executable=sys.executable):
        # type: (str, Optional[ConfigSettings], Union[bytes, Text]) -> None
        """
        Create a project builder.

        :param srcdir: the source directory
        :param config_settings: config settings for the build backend
        :param python_executable: the python executable where the backend lives
        """
        self.srcdir = os.path.abspath(srcdir)  # type: str
        self.config_settings = config_settings if config_settings else {}  # type: ConfigSettings

        spec_file = os.path.join(srcdir, 'pyproject.toml')

        try:
            with open(spec_file) as f:
                self._spec = toml.load(f)
        except FileNotFoundError:
            self._spec = {}
        except PermissionError as e:
            raise BuildException("{}: '{}' ".format(e.strerror, e.filename))
        except toml.decoder.TomlDecodeError as e:
            raise BuildException('Failed to parse pyproject.toml: {} '.format(e))

        _find_typo(self._spec, 'build-system')
        self._build_system = self._spec.get('build-system', _DEFAULT_BACKEND)

        if 'build-backend' not in self._build_system:
            _find_typo(self._build_system, 'build-backend')
            _find_typo(self._build_system, 'requires')
            self._build_system['build-backend'] = _DEFAULT_BACKEND['build-backend']
            self._build_system['requires'] = self._build_system.get('requires', []) + _DEFAULT_BACKEND['requires']

        if 'requires' not in self._build_system:
            raise BuildException("Missing 'build-system.requires' in pyproject.toml")

        self._backend = self._build_system['build-backend']

        self._hook = pep517.wrappers.Pep517HookCaller(
            self.srcdir,
            self._backend,
            backend_path=self._build_system.get('backend-path'),
            python_executable=python_executable,
        )

    @property
    def python_executable(self):  # type: () -> Union[bytes, Text]
        # make mypy happy
        exe = self._hook.python_executable  # type: Union[bytes, Text]
        return exe

    @python_executable.setter
    def python_executable(self, value):  # type: (Union[bytes, Text]) -> None
        self._hook.python_executable = value

    @property
    def build_dependencies(self):  # type: () -> Set[str]
        return set(self._build_system['requires'])

    def get_dependencies(self, distribution):  # type: (str) -> Set[str]
        """
        Returns a set of dependencies
        """
        get_requires = getattr(self._hook, 'get_requires_for_build_{}'.format(distribution))

        try:
            with _working_directory(self.srcdir):
                return set(get_requires(self.config_settings))
        except pep517.wrappers.BackendUnavailable:
            raise BuildException("Backend '{}' is not available.".format(self._backend))
        except Exception as e:  # noqa: E722
            raise BuildBackendException('Backend operation failed: {}'.format(e))

    def check_dependencies(self, distribution):  # type: (str) -> Set[str]
        """
        Returns a set of the missing dependencies

        :param distribution: Distribution to build (sdist or wheel)
        """
        dependencies = self.get_dependencies(distribution)
        dependencies.update(self.build_dependencies)

        return {dep for dep in dependencies if not check_version(dep)}

    def build(self, distribution, outdir):  # type: (str, str) -> None
        """
        Builds a distribution

        :param distribution: Distribution to build (sdist or wheel)
        :param outdir: Output directory
        """
        build = getattr(self._hook, 'build_{}'.format(distribution))
        outdir = os.path.abspath(outdir)

        if os.path.exists(outdir):
            if not os.path.isdir(outdir):
                raise BuildException("Build path '{}' exists and is not a directory".format(outdir))
        else:
            os.mkdir(outdir)

        try:
            with _working_directory(self.srcdir):
                build(outdir, self.config_settings)
        except pep517.wrappers.BackendUnavailable:
            raise BuildException("Backend '{}' is not available.".format(self._backend))
        except Exception as e:  # noqa: E722
            raise BuildBackendException('Backend operation failed: {!r}'.format(e))
